import libdsl.common;
import query;

class FromTree : Tree{
	string var;
	WhereTree where;
	this(Tree varTree, Tree whereTree){
		var = (cast(VariableTree)varTree).name;
		where = cast(WhereTree)whereTree;
	}
	override void ast2ir(Context c){
		c.addArg!(Point[])("args");
		c.defFunc("where", &defWhere);
		c.return_(c=>c.callOpe("from",
							   c.getFunc("where"),
							   c=>c.callOpe("var"~(Point[]).mangleof, c.getLocal("args"))));
	}
	void defWhere(Context c){
		c.addArg!(Point)(var);
		where.ast2ir(c);
	}
}

class WhereTree : Tree{
	Tree cond;
	this(Tree cond){
		this.cond = cond;
	}
	override void ast2ir(Context c){
		c.return_(cond);
	}
}

class CondTree : Tree{
	Tree cond, then, else_;
	this(Tree cond, Tree then, Tree else_){
		this.cond = cond;
		this.then = then;
		this.else_ = else_;
	}
	override void ast2ir(Context c){
		c.callOpe("cond", cond, then, else_);
	}
}

class BinExp : Tree{
	string ope;
	Tree l, r;
	this(Token ope, Tree l, Tree r){
		this.ope = ope.peek!(string);
		this.l = l;
		this.r = r;
	}
	override void ast2ir(Context c){
		c.callOpe("bin"~ope, l, r);
	}
}

class NotTree : Tree{
	Tree tree;
	this(Tree tree){
		this.tree = tree;
	}
	override void ast2ir(Context c){
		c.callOpe("unaryNot", tree);
	}
}
abstract class PostfixTree : Tree{
	Tree base;
}
class PropTree : PostfixTree{
	string var;
	this(string var){ this.var = var; }
	override void ast2ir(Context c){
		c.callOpe("."~var, base);
	}
}

class IntegerTree : Tree{
	int num;
	this(Token num){
		this.num = num.peek!(int);
	}
	override void ast2ir(Context c){
		c.getConstant(num)(c);
	}
}

class VariableTree : Tree{
	string name;
	this(Token name){
		this.name = name.peek!(string);
	}
	override void ast2ir(Context c){
		c.callOpe("var", c.getLocal(name));
	}
}