import libdsl.common;

struct Tokenizer{
	import std.ascii, std.algorithm, std.utf, std.array, std.functional, std.conv;
	static immutable keywords = ["from", "where"];
	static immutable symbols = ["||", "&&", "=="];
	static immutable monosymbols = "?:<+-*/.";
	static immutable symbolchars = sort((symbols.reduce!("a~b")~monosymbols).toUTF32.dup).uniq.array.toUTF8;
	string src;
	Token front_;

	this(string src){
		this.src = src;
		front_ = next;
	}

	bool empty(){
		return front_ is null;
	}
	Token front(){
		return front_;
	}
	void popFront(){
		front_ = next;
	}

	static bool isNameStartDenied(dchar c){
		return c.isDigit || isNameDenied(c);
	}
	static bool isNameDenied(dchar c){
		return c.isWhite || symbolchars.countUntil(c) != -1;
	}
	void skipSpace(){
		import std.stdio;
		auto offset = src.countUntil!((a)=>" \n\t".countUntil(a)==-1);
		if(offset != -1) src = src[offset..$];
	}
	bool startsWithKeyword(string keyword){
		return src.startsWith(keyword) &&
			(src.length == keyword.length || isNameDenied(src[keyword.length]));
	}
	string slice(string str, size_t begin, ptrdiff_t end){
		if(begin == -1){
			assert(end == -1);
			return str[$..$];
		}
		if(end == -1) return str[begin..$];
		return str[begin..end];
	}
	Token next(){
		skipSpace();
		if(src.length == 0) return null;
		
		string tokenize(){
			auto func(){
				foreach(string keyword; keywords)
					if(startsWithKeyword(keyword)) return keyword;
				foreach(string symbol; symbols)
					if(src.startsWith(symbol)) return symbol;
				size_t dummy;
				auto top = src.decode(dummy);
				foreach(symbol; monosymbols)
					if(top == symbol) return [symbol];
				if(!isNameStartDenied(src[0])){
					return slice(src, 0, src.countUntil!(isNameDenied));
				}
				return null;
			}
			auto result = func();
			src = src[result.length..$];
			return result;
		}
		auto stringToken = tokenize();
		if(stringToken != null) return token(stringToken);
		
		if(isDigit(src[0])){
			import std.conv;
			auto tokenOffset = src.countUntil!(not!(isDigit));
			auto value = to!int(slice(src, 0,tokenOffset));
			src = slice(src, tokenOffset,-1);
			return token(value);
		}
		assert(false, `Invalid token:"`~src[0..min($,10)]~`"`);
	}
}
auto tokenize(string src){
	return Tokenizer(src);
}