module libdsl.interpreter;
import std.array;
import std.algorithm;
import std.stdio;
import std.traits;
import std.typecons;
import std.typetuple;
import std.conv;
import std.variant;
public import libdsl.common;
import libdsl.package_common;

enum label : size_t { init=size_t.init }

var_id f2v(func_id f){ return cast(var_id)(-(f+1)); }
func_id v2f(var_id v){ return cast(func_id)(-v-1); }

struct stack{
	size_t[] data;
	@property
	public static size_t elemNum(T)(){
		return elemNum(T.sizeof);
	}
	public static size_t elemNum()(size_t size){
		return (size + typeof(data[0]).sizeof - 1)/typeof(data[0]).sizeof;
	}
	T popRaw(T)(){
		size_t sp = data.length - elemNum(T.sizeof);
		T ret = *cast(T*)(&data[sp]);
		data.length = sp;
//		write(",", ret);
		return ret;
	}
	auto popRawData(){
		const type = popRaw!(TypeHandler);
		auto num = elemNum(type.type.tsize);
		auto sp = data.length - num;
		const val = data[sp..$];
		auto ret = tuple(type, val);
		data.length = sp;
		return ret;
	}
	T pop(T)(){
		auto type = popRaw!(TypeHandler);
		assert(TypeHandler.instance!T is type, "pop as "~T.stringof~" but "~type.type.toString~" on stack");
		return popRaw!T;
	}
	T popAssume(T)(){
		auto data = popRawData();
		auto type = data[0];
		auto ptr = data[1].ptr;
		if(TypeHandler.instance!T is type) return *cast(T*)ptr;
		assert(TypeHandler.instance!T in type.convert, "Can't convert to "~T.stringof~" from "~type.type.toString);
		Unqual!T ret;
		type.convert[TypeHandler.instance!T](ptr, &ret);
		return ret;
	}
	private size_t pushRaw(T)(T value){
		auto prev = data.length;
		data.length += elemNum(T.sizeof);
		*(cast(Unqual!T*)&data[prev]) = value;
		return data.length;
	}
	size_t push(T)(T val){
		pushRaw(val);
		pushRaw(TypeHandler.instance!T);
//		write("->", val);
		return data.length;
	}
	size_t pushRaw()(TypeHandler handler, const typeof(data) arr){
		data ~= arr;
		pushRaw(handler);
		return data.length;
	}
	size_t pushRaw(T)(TypeHandler handler, const(T)[] arr) if(!is(T == typeof(data[0]))){
		auto prev = data.length;
		data.length += elemNum(T.sizeof*arr.length);
		(cast(T*)&data[prev])[0..arr.length] = arr[];
		pushRaw(handler);
		return data.length;
	}

	ref T getLocalVar(T)(size_t id){
		return *cast(T*)(&data[id]);
	}
	string toString(){
		if(data.length == 0) return "[]";
		string ret = "]";
		size_t i = data.length-1;
		while(true){
			auto pi = i;
			auto handler = *cast(TypeHandler*)&data[i];
			i -= elemNum(handler.type.tsize);
			ret = handler.type.toString()~'('~handler.toStringValue(cast(void*)&data[i])~')'~ret;
			i -= elemNum!(TypeHandler*);
			if(i > pi) break;
			ret = ','~ret;
		}
		return '['~ret;
	}
}

union Type{
	string name;
	TypeInfo info;
/+	const hash_t toHash(){
		if(__ctfe)
			return typeid(string;
	}
	const bool opEquals(ref const KeyType s){
	}
	const int opCmp(ref const KeyType s){
	}+/
}

template Expression(alias action_, symbols_...){
	alias action_ action;
	alias symbols_ symbols;
}

template ArgsType(args_...){
	alias args_ args;
}

template DefOpe(string name_, alias func, size_t branches_, RetType_, alias args_, option_...){
	enum name = name_;
	alias func exec;
	enum branches = branches_;
	alias RetType_ RetType;
	alias args_.args args;
	alias option_ option;
}

template Machine(alias funcs){
	template publicFuncs(string name, names...){
		static if(names.length > 0){
			alias publicFuncs!names next;
		}else{
			alias TypeTuple!() next;
		}
		static if(name.startsWith("_args_") || name.startsWith("_branches_")){
			alias next publicFuncs;
		}else{
			mixin("alias funcs."~name~" func;");
			static if(__traits(getProtection, func) == "public"){
				alias TypeTuple!(DefOpe!(name, func, mixin("funcs._branches_"~name),mixin("ArgsType!(funcs._args_"~name~")")), next) publicFuncs;
			}else{
				alias next publicFuncs;
			}
		}
	}
	alias Machine!(publicFuncs!(__traits(derivedMembers, funcs))) Machine;
}

struct ip_t{
	size_t payload;
	alias payload this;
}

struct Machine(funcs...){
	enum embedOpe:size_t{call,vcall,return_,alloc,free,while_,jump,constant}
	template names(alias f, fs...){
		enum name = f.name;
		static if(fs.length > 0){
			alias TypeTuple!(name, names!fs) names;
		}else{
			alias TypeTuple!(name) names;
		}
	}
	template Pair(A_, B_){
		alias A_ A;
		alias B_ B;
	}
	template ConvTypes(T...){
		template Impl(Type, size_t i, P...){
			static if(i == T.length){
				alias P Impl;
			}else static if(!is(Type == T[i]) && is(Type : T[i])){
				alias Impl!(Type, i+1, P, Pair!(Type, T[i])) Impl;
			}else{
				alias Impl!(Type, i+1, P) Impl;
			}
		}
		template Impl2(size_t i, P...){
			static if(i == T.length){
				alias P Impl2;
			}else{
				alias Impl2!(i+1, P, Impl!(T[i], 0)) Impl2;
			}
		}
		alias Impl2!0 ConvTypes;
	}
	alias UsedTypes!(funcs) usedTypes;
	alias ConvTypes!usedTypes convTypes;
	static immutable size_t[string[2]] convPos;
	static this(){
		foreach(i, P; convTypes){
			convPos[[P.A.stringof,P.B.stringof]] = i;
			TypeHandler.addConvert!(P.A, P.B)();
		}
	}
	enum embedOpeLength = embedOpe.max+1+convTypes.length;

	alias size_t code_t;
	static size_t elemNum(T)(){ return elemNum(T.sizeof); }
	static size_t elemNum()(size_t size){ return (size+code_t.sizeof-1)/code_t.sizeof; }
	enum handlerNum = elemNum!(TypeHandler);
	class TranslateContext : Context{
		code_t[][]* codes_;
		@property
		ref code_t[][] codes(){ return *codes_; }
		@property
		ref code_t[] code(){ return codes[proc]; }
		size_t proc;
		size_t allocated[];
		var_id[string][] varPos;
		TranslateContext[string][] funcContext = [null];
		TranslateContext parent;
		TypeHandler prevType;
		TypeHandler returnType;
		const OpeParam[][string] opeParams;

		this(const OpeParam[][string] params){
			opeParams = params;
		}
		this(ref code_t[][] codes){
			this(makeOpeParams!funcs);
			codes_ = &codes;
			proc = 0;
			if(codes.length == 1 && codes[0].length == 0) {
				codes[0] ~= 0;
			}
			allocated = [0];
			varPos = [null];
		}
		this(TranslateContext parent){
			this(parent.codes);
			this.parent = parent;
			codes.length++;
			proc = codes.length-1;
		}

		void defLocalImpl(TypeHandler t, string name){
			varPos[$-1][name] = cast(var_id)allocated[$-1];
			code ~= embedOpe.alloc;
			embedRaw(t);
			allocated[$-1] += elemNum(t.size) + elemNum!(TypeHandler);
		}
		auto getPos(T)(T pos, string name){
			typeof("" in pos[0]) result;
			foreach_reverse(p; pos){
				result = name in p;
				if(result !is null) break;
			}
			assert(result, name~" not found in "~to!string(pos));
			return *result;
		}
		var_id getLocalPos(string name){ return getPos(varPos, name); }
		compilable getLocal(string name){
			auto pos = getLocalPos(name);
			assert(pos >= 0);
			return getConstantT(pos);
		}
		compilable getFunc(string name){
			auto pos = getLocalPos(name);
			assert(pos < 0);
			return getConstantT(pos.v2f);
		}
		TypeHandler getFuncRetType(string name){
			return getPos(funcContext, name).returnType;
		}
		private compilable getConstantT(T)(T val){ return getConstantImpl(TypeHandler.instance!T, token(val)); }
		protected compilable getConstantImpl(TypeHandler t, const Token val){
			return (c){
				auto c_ = cast(TranslateContext)c;
				c_.code ~= embedOpe.constant;
				c_.embed(t, val.getVoidArr);
				c_.prevType = t;
			};
		}
		void defFunc(string name, compilable func){
			auto fid = f2v(cast(func_id)(codes.length));
			auto c = new TranslateContext(this);
			c.varPos[0][name] = varPos[$-1][name] = fid;
			c.funcContext[0][name] = funcContext[$-1][name] = c;
			c.code.length = 1;
			c.code[0] = 0;
			func(c);
		}
		void addArgImpl(TypeHandler t, string name){
			assert(name !in varPos[0]);
			code[0]++;
			code.length += handlerNum;
			for(size_t i=code.length-handlerNum*2; i>0; i-=handlerNum)
				code[i+handlerNum..i+handlerNum*2] = code[i..i+handlerNum];
			embedRaw(t, 1);
			size_t diff = elemNum(t.size) + elemNum!(TypeHandler);
			foreach(ref v; varPos[0]) if(v >= 0) v += diff;
			varPos[0][name] = cast(var_id)0;
			allocated[0] += diff;
		}
		void callFunc(string name, const compilable[] args...){
			foreach(a; args) a(this);
			var_id v = getLocalPos(name);
			if(v < 0){
				auto f = v2f(v);
				code ~= embedOpe.call;
				embedRaw(f);
				prevType = getFuncRetType(name);
			}else{
				code ~= embedOpe.vcall;
				embedRaw(v);
				prevType = null;
			}
		}
		void setReturnType(TypeHandler type){
			assert(returnType is null || returnType is type);
			returnType = type;
		}
		void return_(compilable c){
			c(this);
			code ~= embedOpe.return_;
			setReturnType(prevType);
		}
		void startScope(){
			allocated ~= allocated[$-1];
			varPos.length++;
			funcContext.length++;
		}
		void endScope(){
			if(allocated.length == 1) return;
			auto size = allocated[$-1] - allocated[$-2];
			if(size){
				code ~= embedOpe.free;
				embedRaw(size);
			}
			allocated.length--;
			varPos.length--;
			funcContext.length--;
		}
		void while_(compilable cond, compilable stmt){
			auto whilePos = code.length;
			startScope();
			cond(this);
			endScope();
			code ~= embedOpe.while_;
			auto pos = code.length;
			embedRaw!ptrdiff_t(0);
			startScope();
			stmt(this);
			endScope();
			code ~= embedOpe.jump;
			embedLabel(whilePos);
			embedLabel(code.length, pos);
		}
		void value(TypeHandler t, const void[] v){
			code ~= embedOpe.constant;
			embed(t, v);
		}
		void convertImpl(string A, string B){
			code ~= embedOpe.max + convPos[[A, B]];
		}

		auto embedRaw(T)(T value){ return embedRaw(value, code.length); }
		auto embedRaw(T)(T value, size_t pos){
			size_t vsize = elemNum!T;
			if(pos+vsize > code.length) code.length = pos+vsize;
			*cast(T*)(&code[pos]) = value;
			return vsize;
		}
		auto embedVoidArray(const void[] value, size_t pos){
			size_t vsize = elemNum(value.length);
			if(pos+vsize > code.length) code.length = pos+vsize;
			auto vc = cast(void[])code[pos..$];
			vc[0..value.length] = value[];
			return vsize;
		}

		auto embedType(T)(){ return embedType!T(code.length); }
		auto embedType(T)(size_t pos){
			return embedRaw(TypeHandler.instance!T, pos);
		}
		void embed(T)(T value){ embed(value, code.length); }
		void embed(T)(T value, size_t pos){
			pos += embedType!T(pos);
			embedRaw(value, pos);
		}
		void embed()(TypeHandler t, const void[] val){ embed(t, val, code.length); }
		void embed()(TypeHandler t, const void[] val, size_t pos){
			pos += embedRaw(t, pos);
			embedVoidArray(val, pos);
		}
		void embedLabel(size_t label) { embedLabel(label, code.length); }
		void embedLabel(size_t label, size_t pos){
			embedRaw!ptrdiff_t(cast(ptrdiff_t)label-(pos+elemNum!size_t), pos);
		}

		private template compilableArgs(size_t num){
			static if(num == 0){
				alias TypeTuple!() compilableArgs;
			}else{
				alias TypeTuple!(compilable, compilableArgs!(num-1)) compilableArgs;
			}
		}
		// TODO : 引数の個数でもオーバーロードできるようにする
		void callOpe(string name, const compilable[] tree...){
			assert(name in opeParams, "operator "~name~" is not found");
			auto opes = opeParams[name];
			auto argsNum = opes[0].argTypes.length;
			
			TypeHandler[] argTypes;
			foreach(i; 0..argsNum){
				tree[i](this);
				argTypes ~= prevType;
			}

			auto result = findOverloads!((a,b)=>a==b||a in b.convert)(opes, argTypes);
			assert(result.length > 0, "operator "~name~"("~argTypes.to!string~") is not mached\n"~opes.to!string);
			assert(result.length == 1, "operator "~name~"("~argTypes.to!string~") is mached definitions\n"~opes.to!string);
			auto p = result[0];
			code ~= embedOpeLength+p.id;
			if(p.branches > 0){
				enum size_tNum = elemNum!size_t;
				size_t[] pos = [code.length];
				foreach(i; 0..p.branches+1) embedRaw!ptrdiff_t(0);
				foreach(i, t; tree[p.argTypes.length..$]){
					size_t lpos = pos[0]+size_tNum*(i-p.argTypes.length+1);
					embedLabel(code.length, pos[0]+size_tNum*(i-p.argTypes.length+1));
					startScope();
					if(t) t(this);
					endScope();
					code ~= embedOpe.jump;
					pos ~= code.length;
					embedRaw!ptrdiff_t(0);
				}
				foreach(p_; pos) embedLabel(code.length, p_);
			}
			prevType = p.returnType;
		}
	}

	stack s;
	stack var;
	size_t fp_s[] = [0];
	size_t fp_var[] = [0];
	const code_t[][] codes;
	size_t proc = 0;
	size_t callLevel = 0;
	size_t[string] funcIds;
	const(code_t[]) code(){ return codes[proc]; }
	this(compilable tree){
		code_t[][] codes = [[]];
		auto c = new TranslateContext(codes);
		tree(c);
		this.codes = codes;
		proc = 0;
		foreach(name, value; c.varPos[0]){
			if(value < 0){
				funcIds[name] = v2f(value);
			}
		}
	}
	
	T getConstant(T)(ref ip_t ip){
		T ret = *cast(T*)&code[ip];
		ip += (T.sizeof+code_t.sizeof-1)/code_t.sizeof;
		return ret;
	}
	const(code_t[]) getConstantPtr(ref ip_t ip, size_t size){
		size = (size + code_t.sizeof-1)/code_t.sizeof;
		auto ret = code[ip..ip+size];
		ip += size;
		return ret;
	}
	template RuntimeContext(RetType, option_...){
		alias option_ option;
		size_t setResult(RetType value){
			s.push(value);
			return 0;
		}
		enum cont = 0;
		ref T getLocalVar(T)(var_id v){
			return *cast(T*)&var.data[fp_var[$-1]+v];
		}
		template call(RT){
			RT call(AT...)(func_id f, AT args){
				debug(pope) writeln("\ncall ",f);
				auto p = proc;
				scope(exit) proc = p;
				fp_var ~= var.data.length;
				scope(exit) fp_var.length--;
				fp_s ~= s.data.length;
				scope(exit) fp_s.length--;
				auto level = callLevel;
				scope(exit) callLevel = level;
				callLevel = 0;
				static if(is(RT == void)){
					exec(f, args);
				}else{
					return *(exec(f, args).peek!RT);
				}
			}
		}
	}
	template branchCode(size_t num){
		static if(num == 0){
			alias TypeTuple!() branchCode;
		}else static if(num == 1){
			enum branchCode = num;
		}else{
			alias TypeTuple!(branchCode!(num-1), num) branchCode;
		}
	}

	auto exec(AT...)(string func, AT args){ return exec(funcIds[func], args); }

	auto opCall(AT...)(AT args){ return exec(0, args); }
	R execFor(R, AT...)(AT args){ return *exec(0, args).peek!R; }
	private auto exec(AT...)(size_t funcId, AT funcArgs){
		auto sFp = s.data.length;
		scope(exit) s.data.length = sFp;
		auto varFp = var.data.length;
		scope(exit) var.data.length = varFp;
		proc = funcId;
		foreach(i, T; Reverse!AT){
			auto at = cast(TypeHandler)code[handlerNum*i+1];
			assert(TypeHandler.instance!T == at, "Argument type missmatch. required "~at.type.toString~" but "~T.stringof);
			var.push!T(funcArgs[funcArgs.length-i-1]);
		}
		ip_t ip = ip_t(handlerNum*code[0]+1);
		while(ip < code.length || proc != 0){
			void return_(){
				s.data.length = fp_s[$-1];
				ip = s.pop!ip_t;
				proc = s.pop!size_t;
				var.data.length = fp_var[$-1];
				fp_s.length--; fp_var.length--;
				callLevel--;
			}
			if(ip >= code.length){
				return_();
				if(ip >= code.length) continue;
			}
			debug(pstack) writeln("stack:",s);
			debug(pvar) writeln("var  :",var);
			debug(pope) write("f:",proc," ip:",ip,"\t");
		ope_switch:
			switch(code[ip++]){
			case embedOpe.call:
				auto id = cast(size_t)(getConstant!func_id(ip));
				auto argNum = id == 0? cast(size_t)0: cast(size_t)(codes[id][0]);
				debug(pope) writeln("call (",argNum,")",id);
				fp_var ~= var.data.length;
				foreach(i; 0..argNum) {
					auto d = s.popRawData;
					auto at = cast(TypeHandler)codes[id][handlerNum*(i+1)];
					assert(d[0] == at, "Argument type missmatch. required "~at.type.toString~" but "~d[0].type.toString);
					var.pushRaw(d[0], d[1]);
				}
				s.push(proc);
				s.push(ip);
				fp_s ~= s.data.length;
				proc = id;
				ip = handlerNum*(code[0]+1);
				++callLevel;
				break;
			case embedOpe.vcall:
				auto id = getConstant!var_id(ip);
				debug(pope) writeln("vcall ", id);
				break;
			case embedOpe.return_:
				debug(pope) write("return");
				if(fp_s[$-1] < s.data.length){
					auto v = s.popRawData;
					debug(pope) writeln(v);
					if(callLevel > 0){
						return_();
						s.pushRaw(v[0], v[1]);
					}else{
						return v[0].toVariant(v[1].ptr);
					}
				}else{
					debug(pope) writeln();
					return_();
				}
				break;
			case embedOpe.alloc:
				auto type = getConstant!(TypeHandler)(ip);
				debug(pope) writeln("alloc:",type.type);
				var.pushRaw(type, type.type.init);
				break;
			case embedOpe.free:
				auto num = getConstant!size_t(ip);
				debug(pope) writeln("free ",ip);
				var.data.length -= num;
				break;
			case embedOpe.while_:
				auto dest = getConstant!ptrdiff_t(ip);
				auto cond = s.popAssume!bool;
				debug(pope) writeln("while(",cond, ") -> ",dest);
				if(!cond) ip+=dest;
				break;
			case embedOpe.jump:
				auto dest = getConstant!ptrdiff_t(ip);
				debug(pope) writeln("jump -> ",dest);
				ip += dest;
				break;
			case embedOpe.constant:
				auto type = getConstant!(TypeHandler)(ip);
				auto value = getConstantPtr(ip, type.type.tsize);
				s.pushRaw(type, value);
				debug(pope) writeln("const(",type.type,"):", type.toStringValue(value.ptr));
				break;
				foreach(i,p; convTypes){
				case i+embedOpe.max+1:
					debug(pope) write("cast "~p.A.stringof~"->"~p.B.stringof);
					s.push!(p.B)(s.popAssume!(p.B));
					break ope_switch;
				}
				
				foreach(i,f; funcs){
				case i+embedOpeLength:
				{
					debug(pope) write("ope:(",i,")", names!funcs[i]);
					enum argsName = "args"~i.to!string;
					mixin("f.args "~argsName~";");
					foreach(j, aT; Reverse!(f.args)){
						enum j_ = f.args.length - j - 1;
						mixin(argsName~"[j_] = s.popAssume!aT;");
					}
					auto v = mixin("f.exec!(RuntimeContext!(f.RetType, f.option), "~argsName~", branchCode!(f.branches))()");
					debug(pope) writeln("-> ", v);
					static if(f.branches > 0){
						ip += v;
						auto dest = getConstant!ptrdiff_t(ip);
						ip += dest;
					}
					
					break ope_switch;
				}
				}
				goto default;
			default:
				assert(false, "unknown inst:"~code[ip-1].to!string~" in "~(ip-1).to!string);
			}
		}
		return Variant.init;
	}
}