module libdsl.package_common;
import std.conv;
import std.range;
import std.traits;
import std.typetuple;
import std.typecons;
import std.variant;
import libdsl.common;
import libdsl.parser;

package struct TypeHandler{
	struct Impl{
		TypeInfo type;
		string typename;
		string mangled;
		size_t size;
		bool callable;
		string function(const void*) toStringValue;
		Variant function(const void*) toVariant;
		void function(const void*, void*)[TypeHandler] convert;
		private this(TypeInfo type, string typename, string mangled, size_t size, bool callable, string function(const void*) toStringImpl, Variant function(const void*) toVariantImpl){
			this.type = type;
			this.typename = typename;
			this.mangled = mangled;
			this.size = size;
			this.callable = callable;
			toStringValue = toStringImpl;
			toVariant = toVariantImpl;
		}
		template singleton(T){
			static auto init(){ return Impl(typeid(T),
											T.stringof,
											T.mangleof,
											T.sizeof,
											isCallable!T,
											&toStringImpl,
											&toVariantImpl); }
			static string toStringImpl(const void* ptr){
				import std.conv;
				return to!string(*cast(T*)ptr);
			}
			static Variant toVariantImpl(const void* ptr){
				return Variant(*cast(T*)ptr);
			}
			static Impl instance = init;
		}
	}
	this(size_t s){ haystack = cast(const(Impl)*)s; }
	private this(const(Impl)*ptr){ haystack = ptr; }
	const(Impl)* haystack;
	alias haystack this;
	const string toString(){ return haystack.typename; }
	static void addConvert(T1, T2)(){
		static void impl(const void* s, void* ptr){
			*cast(T2*)ptr = *cast(T1*)s;
		}
		Impl.singleton!T1.instance.convert[TypeHandler.instance!T2] = &impl;
	}
	@property
	static const TypeHandler instance(T)(){
		if(__ctfe){
			Impl* impl = new Impl;
			*impl = Impl.singleton!T.init;
			return TypeHandler(impl);
		}
		static if(is(T == function) || is(T == delegate)){
			return TypeHandler(&Impl.singleton!(SetFunctionAttributes!(T,"D",FunctionAttribute.none)).instance);
		}else{
			return TypeHandler(&Impl.singleton!(Unqual!T).instance);
		}
	}
}
TypeHandler typeHandler(T)(){ return TypeHandler.instance!T; }
const(void)[] toVoidArray(T)(T val){
	return [val];
}

struct OpeParam{
	size_t id;
	size_t branches;
	const TypeHandler[] argTypes;
	TypeHandler returnType;
}
auto makeOpeParams(funcs...)(){
	OpeParam[][string] p;
	foreach(i,f; funcs){
		if(f.name !in p) p[f.name] = [];
		p[f.name] ~= OpeParam(i, f.branches, [staticMap!(TypeHandler.instance, f.args)], TypeHandler.instance!(f.RetType));
	}
	return p;
}
bool checkArgMatch(alias pred)(const TypeHandler[] func, const TypeHandler[] arg){
	foreach(i, f; func){
		if(!pred(f, arg[i])) return false;
	}
	return true;
}
auto findOverloads(alias pred)(const OpeParam[] opes, const TypeHandler[] argTypes){
	const(OpeParam)[] result;
	foreach(o; opes){
		if(checkArgMatch!pred(o.argTypes, argTypes)) result ~= o;
	}
	for(size_t i=0; i<result.length; ++i){
		for(size_t j=0; j<result.length;){
			if(i == j){ ++j; continue; }
			size_t remove;
			if(checkArgMatch!pred(result[j].argTypes, result[i].argTypes)){
				remove = j;
			}else if(checkArgMatch!pred(result[i].argTypes, result[j].argTypes)){
				remove = i;
			}else{
				++j; continue;
			}
			result = result[0..remove]~result[remove+1..$];
			if(remove < i){ /*[0..(j)..i..]*/
				--i;
			}else if(remove == i){ /*[0..(i)..]]*/
				j=0;
			}else{ /*[0..i..(j)..]]*/
			}
		}
	}
	return result;
}

template UniqueTypes(T_...){
	alias T_[0] Args;
	alias T_[1..$] T;
	static if(T.length == 0){
		alias Args UniqueTypes;
	}else static if(is(Args)){
		static if(!is(Args == func_id) && !is(Args == var_id)
				  && staticIndexOf!(Args, T) == -1){
			alias TypeTuple!(T, Args) UniqueTypes;
		}else{
			alias T UniqueTypes;
		}
	}else{
		alias Impl(size_t i : Args.args.length, T...) = T;
		alias Impl(size_t i, T...) = Impl!(i+1, UniqueTypes!(Args.args[i], T));
		static if(is(Args == void)){
			alias T UniqueTypes;
		}else{
			alias Impl!(0, T) UniqueTypes;
		}
	}
}
template UsedTypes(funcs...){
	alias Impl(size_t i : funcs.length, T...) = T;
	alias Impl(size_t i, T...) = Impl!(i+1, UniqueTypes!(funcs[i], UniqueTypes!(funcs[i].RetType, T)));
	alias Impl!0 UsedTypes;
}