module libdsl.parser;
import std.algorithm;
import std.traits;
import std.typetuple;
import std.conv;
import std.range;
import std.typecons;
public import libdsl.common;

size_t getTokenNum(const TokenAA dic, Token t){
	if(t.isTypeOf!eof()) return -1;
	const (size_t)* num;
	//string dicstr = to!string(dic[t]);
	//assert(false, dicstr~"");
	num = t in dic;
	if(num is null) {
		num = t.empty in dic;
		assert(num, to!string(t)~" is not found in terminals.");
	}
	return *num;
}

struct TokenAA{
	size_t[Token] aa;
	size_t[string] ctfeaa;
	
	auto opIndex(Token t){
		if(__ctfe){
			return ctfeaa[t.toString];
		}else{
			return aa[t];
		}
	}
	auto opIndexAssign(size_t value, Token t){
		if(__ctfe){
			return ctfeaa[t.toString] = value;
		}else{
			return aa[t] = value;
		}
	}
	public const auto opBinaryRight(string op)(Token t){
		if(__ctfe){
			return t.toString in ctfeaa;
		}else{
			return t in aa;
		}
	}
}

union TerminalsDictionary{
	Tuple!(void*, size_t)[][string] dic1;
	size_t[Token] dic2;
}

auto createTerminalDictionary(TL...)(){
	TokenAA result;
	foreach(i, T; TL){
		static if(is(T)){
			auto key = token!T(0, 0);
		}else{
			auto key = token(T, 0, 0);
		}
		result[key] = i;
	}
	return result;
}

//enum nonterm : string { init = string.init }
struct nonterm{
	string str;
	string opCast(T : string)(){ return str; }
}
nonterm toNonterm(string name){ return nonterm(name); }
template isNonterm(T){
	enum isNonterm = false;
}
template isNonterm(alias T){
	enum isNonterm = is(typeof(T) == nonterm);
}
template Nonterms(string symbol, T...){
	mixin("enum nonterm "~symbol~"=nonterm(\""~symbol~"\");");
	static if(T.length > 0) mixin Nonterms!T;
}

enum eof : Symbol {init = Symbol.init};

template List(T...){
	alias T list;
}

template Rule(string name_, expressions...){
	enum name = nonterm(name_);
	alias expressions rules;
}

template Expression(alias action_, symbols_...){
	alias action_ action;
	alias symbols_ symbols;
}

template NontermList(rules...){
	alias NontermListImpl!(cast(nonterm[])[], rules[0..$]) NontermList;
}

template NontermListImpl(alias result, rules...){
	static if(rules.length == 0)
		alias result NontermListImpl;
	else static if(countUntil(result, rules[0].name) == -1)
		enum nonterm[] NontermListImpl = NontermListImpl!(result ~ rules[0].name, rules[1..$]);
	else
		enum nonterm[] NontermListImpl = NontermListImpl!(result, rules[1..$]);
}

template TerminalList(rules...){
	alias TerminalListImpl!(List!(), rules) TerminalList;
}
template TerminalListImpl(alias result, rules...){
	static if(rules.length == 0)
		alias result TerminalListImpl;
	else
		alias TerminalListImpl!(TerminalListExpImpl!(result, rules[0].rules), rules[1..$]) TerminalListImpl;
}
template TerminalListExpImpl(alias result, expression...){
	static if(expression.length == 0)
		alias result TerminalListExpImpl;
	else
	alias TerminalListExpImpl!(TerminalListSymbolImpl!(result, expression[0].symbols), expression[1..$]) TerminalListExpImpl;
}
template TerminalListSymbolImpl(alias result, symbols...){
	static if(symbols.length == 0)
		alias result TerminalListSymbolImpl;
	else static if(isNonterm!(symbols[0]) || staticIndexOf!(symbols[0], result.list) != -1)
		alias TerminalListSymbolImpl!(result, symbols[1..$]) TerminalListSymbolImpl;
	else
		alias TerminalListSymbolImpl!(List!(result.list, symbols[0]), symbols[1..$]) TerminalListSymbolImpl;
}

struct Symbol{
	private ptrdiff_t symbol;
	this(ptrdiff_t p){ symbol = p; }
	static start(){ return Symbol(-1); }
	static end(){ return Symbol(0); }
	static void_(){ return Symbol(ptrdiff_t.max); }
	const bool isTerminal(){ return symbol > 0; }
	const bool isNonterminal(){ return symbol < 0; }
}
auto symbol(alias nonterms)(nonterm symbol){
	auto num = nonterms.countUntil(symbol);
	assert(num != -1);
	return Symbol(-num-2);
}
auto symbol(alias tokens, token...)() if(token.length == 1){
	auto num = staticIndexOf!(token[0], tokens.list);
	assert(num != -1);
	return Symbol(num+1);
}

struct Production{
	Symbol head;
	Symbol[] body_;
	invariant(){
		assert(head.isNonterminal);
	}
	const bool isEpsilon(){ return body_.length == 0; }
}

template syntaxTables(alias tokens, alias nonterms, syntax...){
	static if(syntax.length == 1){
		enum syntaxTables = syntaxTable!(tokens, nonterms, symbol!nonterms(syntax[0].name), syntax[0].rules);
	}else{
		enum syntaxTables = syntaxTables!(tokens, nonterms, syntax[0]) ~ syntaxTables!(tokens, nonterms, syntax[1..$]);
	}
}

template syntaxTable(alias tokens, alias nonterms, alias name, rules...){
	static if(rules.length == 1){
		enum syntaxTable = [Production(name, syntaxTableImpl!(tokens, nonterms, rules[0].symbols))];
	}else{
		enum syntaxTable = syntaxTable!(tokens, nonterms, name, rules[0]) ~ syntaxTable!(tokens, nonterms, name, rules[1..$]);
	}
}

template syntaxTableImpl(alias tokens, alias nonterms, expressions...){
	static if(expressions.length == 0)
		enum Symbol[] syntaxTableImpl = [];
	else static if(expressions.length == 1){
		static if(isNonterm!(expressions[0]))
			enum syntaxTableImpl = [symbol!nonterms(expressions[0])];
		else
			enum syntaxTableImpl = [symbol!(tokens, expressions[0])];
	}
	else
		enum syntaxTableImpl = syntaxTableImpl!(tokens, nonterms, expressions[0]) ~ syntaxTableImpl!(tokens, nonterms, expressions[1..$]);
}

auto toOldKernel(const Tuple!(int[2], Symbol[])[][] kernels){
	ptrdiff_t[][][] ret;
	ret.length = kernels.length;
	foreach(i,k; kernels){
		ret[i].length = k.length;
		foreach(j,l; k){
			ret[i][j].length = 2+l[1].length;
			ret[i][j][0..2] = l[0][];
			foreach(j2, l2; l[1]) ret[i][j][j2+2] = l2.symbol;
		}
	}
	return ret;
}

struct LaLrParser(Symbol, SyntaxTreeType, alias nonterms, alias terms, alias exec){
	const Production[] table;
	const ptrdiff_t[] actionTable;
	const ptrdiff_t[] gotoTable;
	const size_t srConflictNum;
	struct Action{
		enum accept = Action(ptrdiff_t.min);
		enum error = Action(0);
		ptrdiff_t action;
		const bool isShift(){ return action > 0; }
		const bool isReduce(){ return action < 0; }
		const size_t ruleNum(){
			assert(isReduce);
			return -action-1;
		}
		static auto reduce(size_t ruleNum){ return Action(-ruleNum-1); }
		static auto shift(size_t state){ return Action(state); }
	}

	this(Production[] table){
		this.table = table;
		const firstSet = first(table);
		const result = makeLalrKernels(table, firstSet);
		_writeln("kernels:",result[0],"\ngotos:",result[1]);
		const lalrKernels = result[0];
		const lalrGoto = result[1];
		size_t conflict;
		actionTable = createAction(firstSet, lalrKernels, lalrGoto, conflict);
		srConflictNum = conflict;
		gotoTable = createGoto(lalrKernels, lalrGoto);
	}
	private auto createAction(const Symbol[][Symbol] firstSet, const LaLrItem[][] lalrKernels, const size_t[GotoIndex] lalrGoto, out size_t srConflictNum){
		auto action = new ptrdiff_t[](lalrKernels.length);
		auto tmp = new Action[](terms.list.length+1);
		auto maxTmp = new ptrdiff_t[](table.length);
		foreach(ptrdiff_t i, i_; lalrKernels){
			tmp[] = Action.error;
			foreach(j; makeLalrClosure(table, firstSet, i_)){
				if(j.isReduction(table)){
					if(j.rule(table).head == Symbol.start)
						tmp[0] = Action.accept;
					else
						foreach(p; j.lookaheads){
							assert(!tmp[p.symbol].isReduce, "reduce-reduce confliction");
							if(!tmp[p.symbol].isShift)
								tmp[p.symbol] = Action.reduce(j.ruleNum);
							else
								//shift-reduce confliction
								++srConflictNum;
						}
				}else{
					auto a = j.dot(table);
					auto g = lalrGoto[GotoIndex(i, a)];
					if(a.isTerminal){
						//shift-reduce confliction
						if(tmp[a.symbol].isReduce) ++srConflictNum;
						assert(!tmp[a.symbol].isShift || tmp[a.symbol] == Action.shift(g), "shift-shift confliction");
						tmp[a.symbol] = Action.shift(g);
					}
				}
			}
			maxTmp[] = 0;
			foreach(j; tmp)
				if(j.isReduce && j != Action.accept) ++maxTmp[j.ruleNum];
			Action maxRed;
			ptrdiff_t maxNum = 0;
			foreach(ptrdiff_t j, j_; maxTmp)
				if(maxNum < j_){
					maxNum = j_;
					maxRed = Action.reduce(j);
				}
			ptrdiff_t[] actionTmp = [];
			foreach(ptrdiff_t j, j_; tmp)
				if(j_.isShift)
					actionTmp ~= [j, j_.action];
				else if(j_.isReduce && j_ != maxRed)
					actionTmp ~= [j, j_.action];
			if(maxNum > 0)
				actionTmp ~= [-1, maxRed.action];
			else
				actionTmp ~= [-1, 0];
			ptrdiff_t position = -1;
			foreach(ptrdiff_t j; 0..i){
				position = action[j];
				ptrdiff_t d = 0;
				while(action[position+d] != -1){
					if(actionTmp[d] != action[position+d]){
						position = -1;
						break;
					}
					d++;
					if(actionTmp[d] != action[position+d]){
						position = -1;
						break;
					}
					d++;
				}
				if(position != -1 && action[position+d] == actionTmp[d] && action[position+d+1] == actionTmp[d+1]){
					action[i] = position;
					break;
				}else
					position = -1;
			}
			if(position == -1){
				action[i] = action.length;
				action ~= actionTmp;
			}
		}
		_writeln("action:",action);
		return action;
	}
	private auto createGoto(const LaLrItem[][] lalrKernels, const size_t[GotoIndex] lalrGoto){
		auto gotoTable = new ptrdiff_t[](nonterms.length);
		auto tmp = new Symbol[](lalrKernels.length);
		auto maxTmp = new ptrdiff_t[](lalrKernels.length);
		_writeln("maxTmp:",maxTmp.length, ",table:",table.length);
		foreach(i, s; nonterms){
			tmp[] = Symbol.start;
			maxTmp[] = 0;
			foreach(ptrdiff_t j; 0..lalrKernels.length){
				auto g = GotoIndex(j, symbol!nonterms(s)) in lalrGoto;
				if(g == null) continue;
				tmp[j] = Symbol(*g);
			}
			foreach(j; tmp)
				if(j != Symbol.start) maxTmp[j.symbol]++;
			ptrdiff_t maxNum, max;
			max = 0;
			foreach(ptrdiff_t j, j_; maxTmp)
				if(max <= j_) {
					max = j_;
					maxNum = j;
				}
			_writeln("GOTO ",-i-2,":",tmp);
			ptrdiff_t[] gotoTmp = [];
			foreach(ptrdiff_t j, j_; tmp){
				if(j_ == Symbol(maxNum) || j_ == Symbol.start) continue;
				gotoTmp ~= [j, j_.symbol];
			}
			gotoTmp ~= [-1, maxNum];
			ptrdiff_t position = -1;
			foreach(ptrdiff_t j; 0..i){
				position = gotoTable[j];
				ptrdiff_t d = 0;
				while(gotoTable[position+d] != -1){
					if(gotoTmp[d] != gotoTable[position+d]){
						position = -1;
						break;
					}
					d++;
					if(gotoTmp[d] != gotoTable[position+d]){
						position = -1;
						break;
					}
					d++;
				}
				if(position != -1 && gotoTable[position+d] == gotoTmp[d] && gotoTable[position+d+1] == gotoTmp[d+1]){
					gotoTable[i] = position;
					break;
				}else
					position = -1;
			}
			if(position == -1){
				gotoTable[i] = gotoTable.length;
				gotoTable ~= gotoTmp;
			}
		}
		_writeln("goto:", gotoTable);
		return gotoTable;
	}

	private auto getAction(size_t state, size_t token){
		token++;
		size_t offset = actionTable[state];
		while(actionTable[offset] != token && actionTable[offset] != -1) offset += 2;
		return actionTable[offset+1];
	}

	private auto getGoto(size_t state, Symbol reduce){
		size_t offset = gotoTable[-reduce.symbol-2];
		while(gotoTable[offset] != state && gotoTable[offset] != -1) offset += 2;
		return gotoTable[offset+1];
	}

	auto parse(T)(T tokenizer){
		Token getToken(){
			if(tokenizer.empty) return token!eof(size_t.max, size_t.max);
			auto t = tokenizer.front;
			tokenizer.popFront;
			return t;
		}
		const terminalDic = createTerminalDictionary!(terms.list)();
		auto token = getToken();
		size_t tokenNum = getTokenNum(terminalDic, token);
		size_t[] stack = [0];
		Object[] params = [];
		while(true){
			auto action = getAction(stack[$-1], tokenNum);
			assert(action != 0, "Syntax error : ("~token.at~"):"~token.toString);
			if(action > 0){
				stack ~= action;
				params ~= token;
				token = getToken();
				tokenNum = getTokenNum(terminalDic, token);
			}else{
				if(table.length <= -action-1) break;
				size_t popNum = table[-action-1].body_.length;
				Object p = exec(-action-1, params[$-popNum..$]);
				params.length -= popNum;
				stack.length -= popNum;
				params ~= p;
				stack ~= getGoto(stack[$-1], table[-action-1].head);
			}
		}
		return cast(SyntaxTreeType)params[$-1];
	}
}

struct LaLrItem{
	Lr0Item item;
	Symbol lookaheads[];
	alias item this;
	const auto dup(){ return LaLrItem(item, lookaheads.dup); }
	void addLookahead(Symbol s){
		if(lookaheads.countUntil(s) == -1) lookaheads ~= s;
	}
}
auto makeLalrKernels(const Production[] table, const Symbol[][Symbol] firstSet){
	auto lr0KernelsResult = makeLr0Kernels(table, table.length-1);
	auto lr0Kernels = lr0KernelsResult[0];
	auto lr0Goto = lr0KernelsResult[1];
	_writeln("LR0 Kernel:",lr0Kernels,"\nLR0 goto:",lr0Goto);
	auto ret = new LaLrItem[][](lr0Kernels.length);
	foreach(i, lr0Kernel; lr0Kernels){
		ret[i].length = lr0Kernel.length;
		foreach(j, kernel; lr0Kernel)
			ret[i][j].item = kernel;
	}
	ptrdiff_t[2][][ptrdiff_t[2]] prop;
	foreach(kernelsNum, kernels; lr0Kernels){
		foreach(num, kernel; kernels){
			auto result = makeLr1Closure(table, firstSet, [Lr1Item(kernel, Symbol.void_)]);
			_writeln("result:",result);
			for(ptrdiff_t j=0; j<result.length; ++j){
//			foreach(j; result){
//			foreach(j; makeLr1Closure!log(table, firstSet, [[kernel[0], kernel[1], ptrdiff_t.max]])){
//				if(table[j[0]].length-1 == j[1]) continue;
				if(result[j].isReduction(table)) continue;
				auto index = GotoIndex(kernelsNum, result[j].dot(table));
				assert(index in lr0Goto, lr0Goto.to!string~"\n"~index.to!string);
				ptrdiff_t aKerNum, aNum;
				aKerNum = lr0Goto[index];
				auto j_ = result[j].shift;
				foreach(ptrdiff_t i, retKernel; ret[aKerNum])
					if(retKernel.rule(table) is j_.rule(table)){
						aNum = i;
						break;
					}
				if(j_.lookahead == Symbol.void_){
					ptrdiff_t[2] index_ = [kernelsNum, num];
					if((index_ in prop) == null)
						prop[index_] = [[aKerNum, aNum]];
					else
						prop[index_] ~= [aKerNum, aNum];
				}else{
					ret[aKerNum][aNum].addLookahead(j_.lookahead);
				}
			}
		}
	}
	ret[0][0].lookaheads ~= Symbol.end;
	bool added = true;
	while(added){
		added = false;
		foreach(ptrdiff_t i, kernels; ret)
			foreach(ptrdiff_t j, kernel; kernels){
				ptrdiff_t[2] index = [i, j];
				if((index in prop) == null) continue;
				foreach(p; prop[index]){
					foreach(add; kernel.lookaheads)
						ret[p[0]][p[1]].addLookahead(add);
				}
			}
	}
	return tuple(ret, lr0Goto);
}

struct Lr0Item{
	size_t ruleNum;
	size_t dotPos;
	auto rule(const Production[] table){
		return table[ruleNum];
	}
	auto dot(const Production[] table){
		return rule(table).body_[dotPos];
	}
	auto isReduction(const Production[] table){
		return rule(table).body_.length == dotPos;
	}
	auto isInitial(){ return dotPos == 0; }
	auto shift(){ return Lr0Item(ruleNum, dotPos+1); }
	auto next(const Production[] table){ return rule(table).body_[dotPos..$]; }
}
struct GotoIndex{
	size_t state;
	Symbol symbol;
}
auto makeLr0Kernels(const Production[] table, size_t startRule){
	Lr0Item[][] ret = [[Lr0Item(startRule, 0)]];
	size_t[GotoIndex] gotoTable;
	bool added = true;
	while(added){
		added = false;
		for(size_t i=0; i<ret.length; ++i){
			auto closure = makeClosure(table, ret[i]);
			_writeln(ret[i]," closure:",closure);
			Symbol[] checkedList = [];
			foreach(j, _j; closure){
				if(_j.isReduction(table)) continue;
				auto x = _j.dot(table);
				bool checked = false;
				foreach(c; checkedList) checked |= (c == x);
				if(checked) continue;
				checkedList ~= x;
				Lr0Item[] kernel = [];
				foreach(k; closure[j..$]){
					if(k.isReduction(table) ||
					   k.dot(table) != x) continue;
					kernel ~= k.shift;
				}
				bool toAdd = true;
				foreach(num, r; ret){
					if(kernel.length == r.length){
						bool exist = false;
						foreach(r_; r){
							exist = false;
							foreach(k; kernel){
								if(r_ == k){
									gotoTable[GotoIndex(i, x)] = num;
									exist = true;
									break;
								}
							}
							if(!exist) break;
						}
						if(exist){
							toAdd = false;
							break;
						}
					}
				}
				if(toAdd){
					gotoTable[GotoIndex(i, x)] = ret.length;
					ret ~= kernel;
				}
			}
		}
	}
	return tuple(ret, gotoTable);
}

auto makeClosure(const Production[] table, Lr0Item[] kernel){
	Lr0Item[] ret = kernel.dup;
	Symbol[] checkedList = [];
	bool added = true;
	while(added){
		added = false;
		for(size_t j=0; j<ret.length; ++j){
			Symbol x;
			if(ret[j].isReduction(table) || !(x=ret[j].dot(table)).isNonterminal) continue;
			bool checked = false;
			foreach(c; checkedList) checked |= (c == x);
			if(checked) continue;
			checkedList ~= x;
			foreach(i, t; table)
				if(t.head == x) ret ~= Lr0Item(i, 0);
		}
	}
	return ret;
}

auto makeLalrClosure(const Production[] table, const Symbol[][Symbol] firstSet, const LaLrItem[] kernel){
	auto ret = new LaLrItem[](kernel.length);
	foreach(i, k; kernel) ret[i] = k.dup;
	for(ptrdiff_t i=0; i<ret.length; ++i){
		if(ret[i].isReduction(table)) continue;
		auto b = ret[i].dot(table);
		if(b.isTerminal) continue;
		Symbol[] first_ = [];
		foreach(f; ret[i].lookaheads)
			foreach(f_; first(firstSet, ret[i].shift.next(table).dup~f))
				if(first_.countUntil(f_) == -1) first_ ~= f_;
		if(first_.length == 0) continue;
		foreach(ruleNum, g; table){
			if(g.head != b) continue;
			ptrdiff_t num = -1;
			foreach(ptrdiff_t i_, r; ret)
				if(r.rule(table) is g && r.isInitial)
					num = i_;
			if(num == -1){
				auto tmp = LaLrItem(Lr0Item(ruleNum, 0), first_);
				if(ret.countUntil(tmp) == -1) ret ~= tmp;
			}else{
				foreach(f; first_)
					ret[num].addLookahead(f);
			}
		}
	}
	return ret;
}

struct Lr1Item{
	Lr0Item item;
	Symbol lookahead;
	alias item this;
	auto shift(){ return Lr1Item(item.shift, lookahead); }
}
auto makeLr1Closure(const Production[] table, const Symbol[][Symbol] firstSet, const Lr1Item[] kernel){
	auto ret = kernel.dup;
	for(ptrdiff_t i=0; i<ret.length; ++i){
		if(ret[i].isReduction(table)) continue;
		auto b = ret[i].dot(table);
		if(b.isTerminal) continue;
		auto first = first(firstSet, ret[i].shift.next(table).dup~ret[i].lookahead);
		foreach(ruleNum, g; table){
			if(g.head != b) continue;
			foreach(b_; first){
				if(b_ == Symbol.end) continue;
				auto tmp = Lr1Item(Lr0Item(ruleNum, 0), b_);
				if(ret.countUntil(tmp) == -1) ret ~= tmp;
			}
		}
	}
	return ret;
}

auto first(const Symbol[][Symbol] first, Symbol[] symbols){
	Symbol[] ret = [];
	bool hasEpsilon;
	foreach(i; symbols){
		hasEpsilon = false;
		if(i.isTerminal)       ret ~= i;
		else if(i.symbol == 0) hasEpsilon = true;
		else if(i.isNonterminal){
			hasEpsilon = false;
			foreach(j; first[i]){
				if(j.symbol == 0) hasEpsilon = true;
				else if(ret.countUntil(j) == -1)
					ret ~= j;
			}
		}
		if(!hasEpsilon) break;
	}
	if(hasEpsilon && ret.countUntil(Symbol.end) == -1) ret ~= Symbol.end;
	return ret;
}

auto first(const Production[] table){
	Symbol[][Symbol] ret;
	bool added = true;
	while(added){
		added = false;
		foreach(rule; table){
			auto x = rule.head;
			if((x in ret) is null) ret[x] = [];
			if(rule.isEpsilon){
				if(ret[x].countUntil(Symbol.end) == -1) {
					ret[x] ~= Symbol.end; added = true;
				}
			}else
				foreach(y; rule.body_){
					if(y.isTerminal){
						if(ret[x].countUntil(y) == -1){
							ret[x] ~= y; added = true;
						}
						break;
					}
					if(y in ret){
						bool hasEpsilon = false;
						foreach(yfirst; ret[y]) {
							if(yfirst == Symbol.end) {
								hasEpsilon = true;
								continue;
							}
							if(ret[x].countUntil(yfirst) == -1){
								ret[x] ~= yfirst; added = true;
							}
						}
						if(!hasEpsilon) break;
					}
				}
		}
	}
	return ret;
}

template selectAct(size_t num){
	auto selectAct(T...)(T args){ return args[num]; }
}
alias bridgeAct = selectAct!0;

template makeExpRules(string name, string base, alias action, alias opePriorities, size_t num = 0){
	import std.conv, std.typecons;
	static if(opePriorities.length == 0){
		alias TypeTuple!(Rule!(name, Expression!(bridgeAct, toNonterm(name~to!string(num)))),
						 Rule!(name~"0", Expression!(bridgeAct, toNonterm(base)))) makeExpRules;
	}else{
		enum thisExp = name~to!string(opePriorities.length);
		enum prevExp = name~to!string(opePriorities.length-1);
		template makeExpression(alias opes){
			static if(opes.length == 0){
				alias Expression!(bridgeAct, toNonterm(prevExp)) makeExpression;
			}else{
				alias TypeTuple!(Expression!(action, toNonterm(thisExp), opes[0], toNonterm(prevExp)), makeExpression!(opes[1..$])) makeExpression;
			}
		}
		alias TypeTuple!(Rule!(thisExp, makeExpression!(opePriorities[0])), makeExpRules!(name, base, action, opePriorities[1..$], num+1)) makeExpRules;
	}
}

void _writeln(T...)(lazy T args){
	import std.stdio;
	debug(parser) if(!__ctfe) std.stdio.writeln(args());
}