module libdsl.syntax;
import libdsl.parser;
import std.algorithm, std.functional, std.typecons, std.uni, std.utf;
public import std.typetuple;

string makeSyntaxRule(string src){
	return "TypeTuple!("~syntax.parse(Tokenizer(src))~")";
}

alias Typedef!(string, "\n", "newline") newline;
alias Typedef!(string, "\n\n", "newlines") newlines;
alias Typedef!(string, "\"\"", "quoted") quotedstr;

class String{
	string s;
	alias s this;
	this(string s){ this.s = s; }
	override string toString(){ return "\""~s~"\""; }
}
auto str(string s){ return new String(s); }
auto str(String s){ return s; }

alias Syntax!(
	"syntax_start".toNonterm, String,
	Rule!("syntax_start",
		  Expression!(bridge, "syntax".toNonterm)),
	Rule!("syntax",
		  Expression!(bridge, "rule".toNonterm),
		  Expression!(cat!(",",1,0,1), "syntax".toNonterm, newlines, "rule".toNonterm)
		),
	Rule!("rule",
		  Expression!(rule, "nonterm".toNonterm, "::", newline, "explist".toNonterm)),
	Rule!("explist",
		  Expression!(bridge, "exp".toNonterm),
		  Expression!(explist, "explist".toNonterm, newline, "exp".toNonterm)
		),
	Rule!("exp",
		  Expression!(exp, "atomlist".toNonterm, "-->", "id".toNonterm),
		  Expression!(expeps, "-->", "id".toNonterm)
		),
	Rule!("atomlist",
		  Expression!(bridge, "atom".toNonterm),
		  Expression!(atomlist, "atomlist".toNonterm, "->", "atom".toNonterm)
		),
	Rule!("atom",
		  Expression!(bridge, "id".toNonterm),
		  Expression!(atomnt, "nonterm".toNonterm),
		  Expression!(convert!quotedstr, quotedstr)),
	Rule!("nonterm",
		  Expression!(at!1, "<", "nonterminable".toNonterm, ">")),
	Rule!("nonterminable",
		  Expression!(convert!string, string),
		  Expression!(cat!(" ",1,string), "nonterminable".toNonterm, string)
		),
	Rule!("id",
		  Expression!(convert!string, string),
		  Expression!(cat!(" ",string,1), string, "postid".toNonterm)),
	Rule!("postid",
		  Expression!(bridge, "postidatom".toNonterm),
		  Expression!(cat!(" ",1,1), "postid".toNonterm, "postidatom".toNonterm)),
		  
	Rule!("postidatom",
		  Expression!(convert!string, string),
		  Expression!(convert!quotedstr, quotedstr),
		  Expression!(convert!string, "<"),
		  Expression!(convert!string, ">")),
	Rule!("spaces",
		  Expression!(nil),
		  Expression!(nil, "spaces".toNonterm, newline),
		  Expression!(nil, "spaces".toNonterm, newlines))) syntax;

String rule(String nt, Token op, Token nl, String exp){
	return ("Rule!(\""~nt~"\","~exp~")").str;
}
String explist(String list, Token nl, String exp){
	return (list~","~exp).str;
}
String exp(String list, Token op, String id){
	return ("Expression!("~id~","~list~")\n").str;
}
String expeps(Token op, String id){
	return ("Expression!("~id~")\n").str;
}
String atomlist(String list, Token op, String atom){
	return (list~","~atom).str;
}
String atomnt(String nt){
	return ("\""~nt~"\".toNonterm").str;
}

String nil(T...)(T args){ return "".str; }
String bridge(String s){ return s; }
template convert(Type, size_t pos = 0){
	String convert(T...)(T t){
		string s = cast(string)(t[pos].peek!Type);
		return s.str;
	}
}
template at(size_t pos){
	String at(T...)(T strs){
		static if(is(T[pos] == Token)){
			return strs[pos].peek!string.str;
		}else{
			return strs[pos];
		}
	}
}
template cat(string sep, Flag...){
	String cat(T...)(T strs){
		string ret;
		foreach(i,s; strs){
			static if(is(typeof(s) == Token)){
				static if(is(Flag[i])){
					string tmp = s.peek!(Flag[i]);
					ret ~= tmp;
					ret ~= sep;
				}
			}else static if(Flag[i]){
				ret ~= s;
				ret ~= sep;
			}
		}
		return ret[0..$-1].str;
	}
}

void skipUntil(alias pred)(ref Tokenizer t){
	ptrdiff_t pos = -1;
	foreach(i, dchar c; t.src){
		if(pred(c)){
			pos = i;
			break;
		}
		if(c == '\n'){
			t.num = 0; t.line++;
		}else{
			t.num++;
		}
	}
	if(pos == -1){
		t.src.length = 0;
	}else{
		t.src = t.src[pos..$];
	}
}
struct Tokenizer{
	// DMD BUGS!
	// immutable string[] keywords = ["::", "->", "-->", "<", ">", "//"];
	string[] keywords = ["::", "->", "-->", "<", ">", "//", "/+"];
	string src;
	Token front_;
	size_t line, num;
	void skipSpace(){
		this.skipUntil!(c=>c=='\n'||!c.isWhite);
		if(src.length == 0) return;
		if(src[0] == '/'){
			if(src[1] == '/'){
				if(src[2] == '/'){
					this.skipUntil!(c=>c=='\n');
					if(src.length){
						src = src[1..$];
						line++; num=0;
					}
					skipSpace();
					return;
				}
				this.skipUntil!(c=>c=='\n');
			}
			if(src[1] == '+'){
				src = src[2..$];
				size_t level = 0;
				while(true){
					this.skipUntil!(c=>c=='+'||c=='/');
					if(src.length == 0) break;
					if(src.startsWith("/+")){
						src = src[2..$];
						++level;
					}else if(src.startsWith("+/")){
						src = src[2..$];
						if(level == 0) break;
						--level;
					}else{
						src = src[1..$];
					}
				}
				skipSpace();
			}
		}
	}
	this(string s){
		src = s;
		this.skipUntil!(not!isWhite);
		if(src.length > 0) popFront;
	}
	bool empty(){
		return front_ is null;
	}
	size_t count = 0;
	Token front(){
		return front_;
	}
	void popFront(){
		if(src.length){
			skipSpace();
			if(src.length){
				if(src[0] == '\n') {
					bool multi = false;
					auto l = line;
					auto n = num;
					while(true){
						src = src[1..$];
						++line; num=0;
						skipSpace();
						if(src.length == 0){
							front_ = null;
							return;
						}
						if(src[0] == '\n'){
							multi = true;
						}else{
							break;
						}
					}
					if(multi){
						front_ = token(newlines.init, l, n);
					}else{
						front_ = token(newline.init, l, n);
					}
					return;
				}
				if(src[0] == '"'){
					auto l = line;
					auto n = num;
					size_t offset = 1;
					++num;
					while(offset < src.length){
						if(src[offset] == '\\'){
							offset += 1+src.stride(offset+1);
							num += 2;
							continue;
						}else if(src[offset] == '"'){
							++offset;
							++num;
							break;
						}
						offset += src.stride(offset);
						++num;
					}
					front_ = token(quotedstr(src[0..offset]), l, n);
					src = src[offset..$];
					return;
				}
				foreach(k; keywords){
					if(src.startsWith(k)){
						front_ = token(k, line, num);
						src = src[k.length..$];
						num += k.length;
						return;
					}
				}
				size_t offset = src.stride;
				auto l = line;
				auto n = num;
			loop:
				while(offset < src.length){
					auto proc = src[offset..$];
					++num;
					foreach(k; keywords){
						if(proc.startsWith(k)) break loop;
					}
					if(proc[0].isWhite || proc[0] == '"') break;
					offset += src.stride(offset);
				}
				front_ = token(src[0..offset], l, n);
				src = src[offset..$];
				return;
			}
		}
		front_ = null;
	}
}