import libdsl.parser;
import libdsl.common;
import libdsl.syntax;
import lexer;
import ast;

alias syntax = Syntax!("query".toNonterm, Tree,
					   makeExpRules!("exp", "factor", binexp, [["||"], ["&&"], ["==","<"], ["+","-"], ["*","/"]]),
					   mixin(makeSyntaxRule(`
<query> ::
	"from" -> <variable> -> <where> --> from

<where> ::
	"where" -> <condexp> --> where

<condexp> ::
	<exp> --> bridgeAct
	<exp> -> "?" -> <condexp> -> ":" -> <condexp> --> condexp

<factor> ::
	"(" -> <condexp> -> ")" --> selectAct!(1)
	"!" -> <factor> --> notexp
	<factor> -> <postfix> --> postfixexp
	<integer> --> bridgeAct
	<variable> --> bridgeAct

<postfix> ::
	"." -> string --> propexp

<integer> ::
	int --> integer

<variable> ::
	string --> variable`)));


Tree from(Token fromToken, Tree variable, Tree where){
	return new FromTree(variable, where);
}
Tree where(Token whereToken, Tree condexp){
	return new WhereTree(condexp);
}
Tree condexp(Tree cond, Token qToken, Tree then, Token colonToken, Tree else_){
	return new CondTree(cond, then, else_);
}
Tree binexp(Tree left, Token opeToken, Tree right){
	return new BinExp(opeToken, left, right);
}
Tree notexp(Token notToken, Tree tree){
	return new NotTree(tree);
}
Tree postfixexp(Tree base, Tree post){
	(cast(PostfixTree)post).base = base;
	return post;
}
Tree propexp(Token dot, Token name){
	return new PropTree(name.peek!string);
}
Tree integer(Token integer){
	return new IntegerTree(integer);
}
Tree variable(Token name){
	return new VariableTree(name);;
}

auto parse(string src){
	return syntax.parse(src.tokenize);
}