import std.typecons;
import libdsl.interpreter;
import libdsl.compiler;
import parser;

struct Point{
	int x, y;
}

auto runDsl(string src, Point[] array){
	auto tree = parse(src);
	auto machine = QueryMachine(tree);
	return machine.execFor!(Point[])(array);
}

auto runDsl(string src)(Point[] array){
	static compilable getTree(){ return parse(src); }
	auto machine = QueryCompiler!(getTree).init;
	return machine.execFor!(Point[])(array);
}


alias TypeTuple!(DefOpe!("from", from, 0, Point[], ArgsType!(func_id, Point[])),
				 DefOpe!("cond", opeCond, 2, Point, ArgsType!bool),
				 DefOpe!("bin||", opeOr, 1, bool, ArgsType!bool),
				 DefOpe!("bin&&", opeAnd, 1, bool, ArgsType!bool),
				 DefBinOpe!("=="), DefBinOpe!("<"),
				 DefBinOpe!("+"), DefBinOpe!("-"), DefBinOpe!("*"), DefBinOpe!("/"),
				 DefOpe!("unary!", unaryNot, 0, bool, ArgsType!bool),
				 DefOpe!(".x", prop, 0, int, ArgsType!Point, "x"),
				 DefOpe!(".y", prop, 0, int, ArgsType!Point, "y"),
				 DefOpe!("var", var, 0, Point, ArgsType!(var_id), Point),
				 DefVar!(Point[]),
				 DefVar!(int delegate(ref Point))) queryOpes;

alias Machine!queryOpes QueryMachine;
alias Compiler!queryOpes QueryCompiler;

auto from(alias context, alias where, alias array)(){
	alias boolCall = context.call!bool;
	Point[] result;
	foreach(i; array) if(boolCall(where, i)) result ~= i;
	return context.setResult(result);
}

auto opeCond(alias context, alias c, alias then, alias else_)(){
	if(c){
		return then;
	}else{
		return else_;
	}
}

auto opeOr(alias context, alias cond, alias next)(){
	if(cond){
		return context.setResult(true);
	}else{
		return next;
	}
}

auto opeAnd(alias context, alias cond, alias next)(){
	if(cond){
		return next;
	}else{
		return context.setResult(false);
	}
}

alias DefBinOpe(string ope) = DefOpe!("bin"~ope, binope, 0, typeof(mixin("0 "~ope~" 0")), ArgsType!(int, int), ope);

auto binope(alias context, alias l, alias r)(){
	return context.setResult(mixin("l "~context.option[0]~" r"));
}

auto unaryNot(alias context, alias cond)(){
	return context.setResult(!cond);
}

auto prop(alias context, alias base)(){
	return context.setResult(mixin("base."~context.option[0]));
}

alias DefVar(T) = DefOpe!("var"~T.mangleof, var, 0, T, ArgsType!(var_id), T);

auto var(alias context, alias pos)(){
	return context.setResult(context.getLocalVar!(context.option[0])(pos));
}